"""
Validation module

Contains routines for input validation of both the masked_norm and
affine_masked_norm procedures.
"""

from __future__ import annotations

from torch import Tensor
from torch import Size as Shape
from torch import float32, float64, float16, bfloat16
from torch import float8_e4m3fn, float8_e5m2
from torch import bool as bool_dtype


# PyTorch doesn't offer abstract type hinting for xfloat* dtypes
# you can work around this by verifying membership on a set of xfloat* dtypes
float_dtype = {
    float64,
    float32,
    float16,
    bfloat16,
    float8_e4m3fn,
    float8_e5m2
}


def validate_masked_norm(inpt: Tensor, mask: Optional[Tensor]) -> None:
    """
    Validates the input of the masked_norm procedure.
    """

    if not inpt.dtype in float_dtype:
        raise ValueError("'inpt' must be a tensor of xfloat* dtype")

    if mask is None:
        return

    if not mask.dtype == bool_dtype:
        # mask tensor must be necessarily of bool dtype
        raise ValueError("'mask' must be a tensor of bool dtype")

    if not len(inpt.shape) > len(mask.shape):
        raise ValueError(
            "the shape of 'inpt' must be larger than the shape of 'mask'"
        )

    # mask shape must partially match input shape
    for axis, dim in enumerate(zip(inpt.shape, mask.shape)):
        inpt_dim, mask_dim = dim
        if inpt_dim != mask_dim:
            raise ValueError(
                f"dim mismatch between 'inpt' and 'mask' on axis {axis}"
        )


def validate_affine_masked_norm(
    inpt: Tensor,
    mask: Optional[Tensor],
    weight: Tensor,
    bias: Optional[Tensor]
) -> None:
    """
    Validates the input of the affine_masked_norm procedure.
    """

    validate_masked_norm(inpt, mask)

    # weight and bias must have proper dtype
    if not weight.dtype in float_dtype:
        raise ValueError("'weight' must be a tensor of xfloat* dtype")

    if not bias is None and not bias.dtype in float_dtype:
        raise ValueError("'bias' must be a tensor of xfloat* dtype")


    # if no mask is passed, weight and bias tensor must have shape
    # inpt.shape[0: -1]
    if mask is None:
        shape = inpt.shape[0: -1]

        if weight.shape != shape:
            raise ValueError(
                f"'weight' must be a tensor with shape {shape}"
            )

        if not bias is None and bias.shape != shape:
            raise ValueError(
                f"'bias' must be a tensor with shape {shape}"
            )

        return

    # if a mask is passed, weight and bias must have the same shape as mask
    if weight.shape != mask.shape:
        raise ValueError(
            "'weight' must have the same shape as 'mask'"
        )

    if not bias is None and bias.shape != mask.shape:
        raise ValueError(
            "'bias' must have the same shape as 'mask'"
        )




"""
Utilities module
"""

from torch import Tensor


def unsqueeze_as(dst: Tensor, src: Tensor) -> Tensor:
    """
    Unsqueezes singleton dim axis according to the shape of a source tensor.
    """

    diff = len(src.shape) - len(dst.shape)

    if diff < 0:
        raise ValueError(
            "the shape of 'src' cannot be smaller than the shape of 'dst'"
        )

    for i in range(diff):
        dst = dst.unsqueeze(-1)

    return dst


def get_factory_key(key: str) -> bool:
    """
    Filtering procedure for selecting factory kwargs.
    """

    if key == "device":
        return True

    if key == "dtype":
        return True

    return False



"""
Nns module testing
"""

from __future__ import annotations

from math import sqrt

from torch import Tensor, tensor, ones_like, zeros_like, ones
from torch.nn import BatchNorm1d, LayerNorm, InstanceNorm1d, GroupNorm

from masked_norm import masked_norm, batched_masked_norm
from masked_norm import MaskedNorm
from masked_norm import LazyMaskedNorm

from masked_norm import affine_masked_norm
from masked_norm import batched_affine_masked_norm
from masked_norm import LazyAffineMaskedNorm
from masked_norm.validation import validate_affine_masked_norm

from .tensor_test_case import TensorTestCase


class TestMaskedNorm(TensorTestCase):
    """
    'MaskedNorm' testing unit
    """

    def assertApprox(
        self: TestMaskedNorm,
        x: Tensor,
        y: Tensor,
        msg: str | None = None
    ) -> None:
        """
        Element-wise approximation assertion method
        """

        cond = (x - y).abs() < self.eps
        cond = cond.all()

        self.assertTrue(cond, msg)

    def setUp(self: TestMaskedNorm) -> None:
        """
        """

        self.inpt_1 = tensor(
            [
                [0.00, 1.00, 2.00],
                [3.00, 4.00, 5.00]
            ]
        )

        self.inpt_2 = tensor(
            [
                [
                    [0.00, 1.00, 2.00],
                    [3.00, 4.00, 5.00],
                    [6.00, 7.00, 8.00],
                    [9.00, 10.0, 11.0]
                ],
                [
                    [12.0, 13.0, 14.0],
                    [15.0, 16.0, 17.0],
                    [18.0, 19.0, 20.0],
                    [21.0, 22.0, 23.0]
                ]
            ]
        )

        self.eps = 1e-6

    def test_forward_mask_0(self: TestMaskedNorm) -> None:
        """
        """

        mask_0 = None

        layer = MaskedNorm()
        layer_out = layer(self.inpt_1, mask_0)

        functional_out = masked_norm(self.inpt_1, mask_0)

        self.assertEqTensor(layer_out, functional_out)

    def test_forward_mask_1(self: TestMaskedNorm) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        layer = MaskedNorm()
        layer_out = layer(self.inpt_1, mask_1)

        functional_out = masked_norm(self.inpt_1, mask_1)

        self.assertEqTensor(layer_out, functional_out)

    def test_forward_batched_mask_0(self: TestMaskedNorm) -> None:
        """
        """

        mask_0 = None

        layer = MaskedNorm(batched=True)
        layer_out = layer(self.inpt_1, mask_0)

        functional_out = batched_masked_norm(self.inpt_1, mask_0)

        self.assertEqTensor(layer_out, functional_out)

    def test_forward_batched_mask_1(self: TestMaskedNorm) -> None:
        """
        """

        mask_2 = tensor(
            [True, True, False]
        )

        layer = MaskedNorm(batched=True)
        layer_out = layer(self.inpt_1, mask_2)

        functional_out = batched_masked_norm(self.inpt_1, mask_2)

        self.assertEqTensor(layer_out, functional_out)

    def test_batch_norm(self: TestMaskedNorm) -> None:
        """
        Asserts the equivalency between the MaskedNorm and BatchNorm
        implementations
        """

        masked_norm_layer = MaskedNorm()
        batch_norm_layer = BatchNorm1d(4, eps=0.0, affine=False)

        inpt = self.inpt_2.transpose(0, 1)
        shape = inpt.shape
        inpt = inpt.flatten(start_dim=1)
        masked_norm_out = masked_norm_layer(inpt)
        masked_norm_out = masked_norm_out.view(*shape)
        masked_norm_out = masked_norm_out.transpose(0, 1)

        batch_norm_out = batch_norm_layer(self.inpt_2)

        # variance biasness bessel correction
        batch_norm_out *= sqrt(5/6)

        self.assertApprox(masked_norm_out, batch_norm_out)

    def test_layer_norm(self: TestMaskedNorm) -> None:
        """
        Asserts the equivalency between the MaskedNorm and LayerNorm
        implementations
        """

        masked_norm_layer = MaskedNorm()
        layer_norm_layer = LayerNorm((4, 3), eps=0.0, elementwise_affine=False)

        inpt = self.inpt_2
        shape = inpt.shape
        inpt = inpt.flatten(start_dim=1)
        masked_norm_out = masked_norm_layer(inpt)
        masked_norm_out = masked_norm_out.view(*shape)

        layer_norm_out = layer_norm_layer(self.inpt_2)

        # variance biassness bessel correction
        layer_norm_out *= sqrt(11/12)

        self.assertApprox(masked_norm_out, layer_norm_out)

    def test_instance_norm(self: TestMaskedNorm) -> None:
        """
        Asserts the equivalency between the MaskedNorm and InstanceNorm
        implementations
        """

        masked_norm_layer = MaskedNorm()
        inst_norm_layer = InstanceNorm1d(4, eps=0.0, affine=False)

        mask_shape = self.inpt_2.shape[0: 2]
        mask = ones(mask_shape).to(bool)

        masked_norm_out = masked_norm_layer(self.inpt_2, mask)

        inst_norm_out = inst_norm_layer(self.inpt_2)

        # variance biassness bessel correction
        inst_norm_out *= sqrt(2/3)

        self.assertApprox(masked_norm_out, inst_norm_out)

    def test_group_norm(self: TestMaskedNorm) -> None:
        """
        Asserts the equivalency between the MaskedNorm and GroupNorm
        implementations
        """

        n_channels = 4
        n_groups = 2

        masked_norm_layer = MaskedNorm()
        group_norm_layer = GroupNorm(
            n_groups,
            n_channels,
            eps=0.0,
            affine=False
        )

        inpt = self.inpt_2.unfold(1, n_groups, n_channels // n_groups)

        inpt = inpt.transpose(2, 3)
        inpt = inpt.flatten(start_dim=0, end_dim=1)

        shape = inpt.shape
        inpt = inpt.flatten(start_dim=1)
        masked_norm_out = masked_norm_layer(inpt)
        masked_norm_out = masked_norm_out.view(*shape)
        masked_norm_out = masked_norm_out.reshape((2, 4, 3))

        group_norm_out = group_norm_layer(self.inpt_2)

        # variance biassness bessel correction
        group_norm_out *= sqrt(5/6)

        self.assertApprox(masked_norm_out, group_norm_out)


class TestLazyMaskedNorm(TensorTestCase):
    """
    'LazyMaskedNorm' testing unit
    """

    def setUp(self: TestLazyMaskedNorm) -> None:
        """
        """

        self.inpt = tensor(
            [
                [0.00, 1.00, 2.00],
                [3.00, 4.00, 5.00]
            ]
        )

    def test_init_mask_0(self: TestLazyMaskedNorm) -> None:
        """
        """

        mask_0 = None

        try:
            layer = LazyMaskedNorm()
            layer.forward(self.inpt, mask_0)
            self.assertTrue(True)

        except:
            self.assertTrue(False)

    def test_init_mask_1(self: TestLazyMaskedNorm) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        try:
            layer = LazyMaskedNorm()
            layer.forward(self.inpt, mask_1)
            self.assertTrue(True)

        except:
            self.assertTrue(False)


class TestLazyAffineMaskedNorm(TensorTestCase):
    """
    'LazyAffineMaskedNorm' testing unit
    """

    def setUp(self: TestAffineMaskedNorm) -> None:
        """
        """

        self.inpt = tensor(
            [
                [0.00, 1.00, 2.00],
                [3.00, 4.00, 5.00]
            ]
        )

    def test_param_init_mask_0(self: TestLazyAffineMaskedNorm) -> None:
        """
        """

        mask_0 = None

        layer = LazyAffineMaskedNorm()

        layer.forward(self.inpt, mask_0)

        try:
            validate_affine_masked_norm(
                self.inpt,
                mask_0,
                layer.weight,
                layer.bias
            )

            self.assertTrue(True)

        except:
            self.assertTrue(False)

    def test_param_init_mask_1(self: TestLazyAffineMaskedNorm) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        layer = LazyAffineMaskedNorm()

        layer.forward(self.inpt, mask_1)

        try:
            validate_affine_masked_norm(
                self.inpt,
                mask_1,
                layer.weight,
                layer.bias
            )
            self.assertTrue(True)

        except:
            self.assertTrue(False)

        weight_1 = ones_like(mask_1)
        bias_1 = zeros_like(mask_1)

        self.assertEqTensor(layer.weight, weight_1)
        self.assertEqTensor(layer.bias, bias_1)

    def test_forward_mask_0(self: TestLazyAffineMaskedNorm) -> None:
        """
        """

        mask_0 = None

        layer = LazyAffineMaskedNorm()

        layer_out = layer(self.inpt, mask_0)

        functional_out = affine_masked_norm(
            self.inpt,
            mask_0,
            layer.weight,
            layer.bias
        )

        self.assertEqTensor(layer_out, functional_out)

    def test_forward_mask_1(self: TestLazyAffineMaskedNorm) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        layer = LazyAffineMaskedNorm()

        layer_out = layer(self.inpt, mask_1)

        functional_out = affine_masked_norm(
            self.inpt,
            mask_1,
            layer.weight,
            layer.bias
        )

        self.assertEqTensor(layer_out, functional_out)

    def test_forward_batched_mask_0(self: TestLazyAffineMaskedNorm) -> None:
        """
        """

        mask_0 = None

        layer = LazyAffineMaskedNorm(batched=True)

        layer_out = layer(self.inpt, mask_0)

        functional_out = batched_affine_masked_norm(
            self.inpt,
            mask_0,
            layer.weight,
            layer.bias
        )

        self.assertEqTensor(layer_out, functional_out)

    def test_forward_batched_mask_1(self: TestLazyAffineMaskedNorm) -> None:
        """
        """

        mask_1 = tensor(
            [True, True, False]
        )

        layer = LazyAffineMaskedNorm(batched=True)

        layer_out = layer(self.inpt, mask_1)

        functional_out = batched_affine_masked_norm(
            self.inpt,
            mask_1,
            layer.weight,
            layer.bias
        )

        self.assertEqTensor(layer_out, functional_out)



"""
Functional module

Contains the functional implementation of masked normalization.
"""

from __future__ import annotations
from typing import Optional

from torch import Tensor, permute

from .validation import validate_masked_norm
from .validation import validate_affine_masked_norm
from .util import unsqueeze_as


def masked_norm(
    inpt: Tensor,
    mask: Optional[Tensor] = None,
) -> Tensor:
    """
    Masked normalization procedure.

    Normalizes elements of the input specified by a mask. The normalization is
    performed along the last axis. If a selection of samples yields null
    variance, the normalization over those elements is ignored, and the values
    are passed along unaltered.
    """

    validate_masked_norm(inpt, mask)

    *_, n = list(inpt.shape)

    if n > 1:
        var = inpt.var(dim=-1, keepdim=True)
    else:
        return inpt

    mean = inpt.mean(dim=-1, keepdim=True)

    var_mask = (var != 0.0).squeeze(-1)
    if mask is None:
        mask = var_mask
    else:
        mask = unsqueeze_as(mask, var_mask)
        mask |= var_mask

    norm = inpt[mask] - mean[mask]
    norm = norm / var[mask].sqrt()

    clone = inpt.clone()
    clone[mask] = norm

    return clone


def affine_masked_norm(
    inpt: Tensor,
    mask: Optional[Tensor],
    weight: Tensor,
    bias: Optional[Tensor],
) -> Tensor:
    """
    Affine masked normalization procedure.

    Normalizes elements of the input specified by a mask. The normalization is
    performed along the last axis. If a selection of samples yields null
    variance, the normalization over those elements is ignored, and the values
    are passed along unaltered. After normalization an affine transformation
    is applied along the normalized axis.
    """

    validate_affine_masked_norm(inpt, mask, weight, bias)

    inpt = masked_norm(inpt, mask)

    inpt = weight[..., None] * inpt

    if bias is not None:
        inpt = inpt + bias[..., None]

    return inpt


def batched_masked_norm(inpt: Tensor, mask: Optional[Tensor]) -> Tensor:
    """
    Batched masked normalization procedure.

    Normalizes elements of the input specified by a mask. The normalization is
    performed along the first axis. If a selection of samples yields null
    variance, the normalization over those elements is ignored, and the values
    are passed along unaltered. Standard masked normalization is performed
    along the last axis, which is usually populated with features. Batched
    masked normalization reverses this order by normalizing along the first
    axis.
    """

    shape = inpt.shape
    n = len(shape)

    perm = list(range(1, n)) + [0]
    inpt = permute(inpt, perm)

    inpt = masked_norm(inpt, mask)

    inv_perm = [n - 1] + list(range(0, n - 1))
    inpt = permute(inpt, inv_perm)

    return inpt


def batched_affine_masked_norm(
    inpt: Tensor,
    mask: Optional[Tensor],
    weight: Tensor,
    bias: Optional[Tensor],
) -> Tensor:
    """
    Batched affine masked normalization procedure.

    Normalizes elements of the input specified by a mask. The normalization is
    performed along the first axis. If a selection of samples yields null
    variance, the normalization over those elements is ignored, and the values
    are passed along unaltered. After normalization an affine transformation
    is applied along the normalized axis. Standard masked normalization is
    performed along the last axis, which is usually populated with features.
    Batched masked normalization reverses this order by normalizing along the
    first axis.
    """

    shape = inpt.shape
    n = len(shape)

    perm = list(range(1, n)) + [0]
    inpt = permute(inpt, perm)

    validate_affine_masked_norm(inpt, mask, weight, bias)

    inpt = masked_norm(inpt, mask)

    inpt = weight[..., None] * inpt

    if bias is not None:
        inpt = inpt + bias[..., None]

    inv_perm = [n - 1] + list(range(0, n - 1))
    inpt = permute(inpt, inv_perm)

    return inpt



"""
Validation module testing
"""

from __future__ import annotations

from torch import Tensor, tensor
from masked_norm.validation import validate_masked_norm
from masked_norm.validation import validate_affine_masked_norm

from .tensor_test_case import TensorTestCase


class TestValidateMaskedNorm(TensorTestCase):
    """
    'validate_masked_norm' testing unit
    """

    def setUp(self: TestValidateMaskedNorm) -> None:
        """
        """

        self.inpt = tensor(
            [
                [0.00, 1.00, 2.00],
                [3.00, 4.00, 5.00]
            ]
        )

    def test_mask_0(self: TestValidateMaskedNorm) -> None:
        """
        """

        mask_0 = None

        try:
            validate_masked_norm(self.inpt, mask_0)
            self.assertTrue(True)

        except:
            self.assertTrue(False)

    def test_mask_1(self: TestValidateMaskedNorm) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        try:
            validate_masked_norm(self.inpt, mask_1)
            self.assertTrue(True)

        except:
            self.assertTrue(False)

    def test_mask_3(self: TestValidateMaskedNorm) -> None:
        """
        """

        mask_3 = tensor(
            [
                [
                    [True]
                ]
            ]
        )

        with self.assertRaisesRegex(
            ValueError,
            "the shape of 'inpt' must be larger than the shape of 'mask'"
        ):
            validate_masked_norm(self.inpt, mask_3)

    def test_mask_4(self: TestValidateMaskedNorm) -> None:
        """
        """

        mask_4 = tensor(
            [True]
        )

        with self.assertRaisesRegex(
            ValueError,
            "dim mismatch between 'inpt' and 'mask'"
        ):

            validate_masked_norm(self.inpt, mask_4)


class TestValidateAffineMaskedNorm(TensorTestCase):
    """
    'validate_affine_masked_norm' testing unit
    """

    def setUp(self: TestValidateAffineMaskedNorm) -> None:
        """
        """


        self.inpt = tensor(
            [
                [0.00, 1.00, 2.00],
                [3.00, 4.00, 5.00]
            ]
        )

    def test_mask_0_weight_1_bias_0(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_0 = None

        weight_1 = tensor(
            [1.0, 1.0]
        )

        bias_0 = None

        try:
            validate_affine_masked_norm(self.inpt, mask_0, weight_1, bias_0)
            self.assertTrue(True)

        except:
            self.assertTrue(False)

    def test_mask_0_weight_1_bias_1(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_0 = None

        weight_1 = tensor(
            [1.0, 1.0]
        )

        bias_1 = tensor(
            [0.0, 0.0]
        )

        try:
            validate_affine_masked_norm(self.inpt, mask_0, weight_1, bias_1)
            self.assertTrue(True)

        except:
            self.assertTrue(False)

    def test_mask_0_weight_2_bias_0(
            self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_0 = None

        weight_2 = tensor(
            [1.0, 1.0, 1.0]
        )

        bias_0 = None

        with self.assertRaisesRegex(
            ValueError,
            "'weight' must be a tensor with shape"
        ):
            validate_affine_masked_norm(self.inpt, mask_0, weight_2, bias_0)

    def test_mask_0_weigth_1_bias_2(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_0 = None

        weight_1 = tensor(
            [1.0, 1.0]
        )

        bias_2 = tensor(
            [0.0, 0.0, 0.0]
        )

        with self.assertRaisesRegex(
            ValueError,
            "'bias' must be a tensor with shape"
        ):
            validate_affine_masked_norm(self.inpt, mask_0, weight_1, bias_2)

    def test_mask_1_weight_1_bias_0(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        weight_1 = tensor(
            [1.0, 1.0]
        )

        bias_0 = None

        try:
            validate_affine_masked_norm(self.inpt, mask_1, weight_1, bias_0)
            self.assertTrue(True)

        except:
            self.assertTrue(False)

    def test_mask_1_weight_1_bias_1(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        weight_1 = tensor(
            [1.0, 1.0]
        )

        bias_1 = tensor(
            [0.0, 0.0]
        )

        try:
            validate_affine_masked_norm(self.inpt, mask_1, weight_1, bias_1)
            self.assertTrue(True)

        except:
            self.assertTrue(False)

    def test_mask_1_weight_2_bias_0(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        weight_2 = tensor(
            [1.0, 1.0, 1.0]
        )

        bias_0 = None

        with self.assertRaisesRegex(
            ValueError,
            "'weight' must have the same shape as 'mask'"
        ):
            validate_affine_masked_norm(self.inpt, mask_1, weight_2, bias_0)

    def test_mask_1_weight_1_bias_2(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        weight_1 = tensor(
            [1.0, 1.0]
        )

        bias_2 = tensor(
            [0.0, 0.0, 0.0]
        )

        with self.assertRaisesRegex(
            ValueError,
            "'bias' must have the same shape as 'mask'"
        ):
            validate_affine_masked_norm(self.inpt, mask_1, weight_1, bias_2)
